@@ -4,7 +4,7 @@ import asyncio
 import json
 import signal
 from concurrent.futures import ThreadPoolExecutor
-from typing import Awaitable, Callable, Optional, Set, Union
+from typing import Any, Awaitable, Callable, Optional, Set, Union
 
 import httpx
 from aiosseclient import aiosseclient
@@ -25,6 +25,8 @@ from .models.executions import (
 # --------------------------------------------------------------------------- #
 
 EVENT_STREAMING_ENDPOINT = "{base}/{organization_id}/events"
+_MAX_RETRIES = 3
+_RETRY_DELAY = 1  # seconds
 ExecutionRequestHandler = Union[
     Callable[[AgentExecution], AgentExecutionResult],
     Callable[[AgentExecution], Awaitable[AgentExecutionResult]],
@@ -47,6 +49,8 @@ def _should_pass_org_id(base_url: Optional[str]) -> bool:
 
 
 class XpanderEventListener:
+    """Listen to Xpander Agent events with built‑in retry logic (SSE + HTTP)."""
+
     # --------------------------------------------------------------------- #
     # Construction                                                          #
     # --------------------------------------------------------------------- #
@@ -61,8 +65,11 @@ class XpanderEventListener:
         should_reset_cache: bool = False,
         with_metrics_report: bool = False,
         max_sync_workers: int = 4,
+        max_retries: int = _MAX_RETRIES,
+        retry_delay: float = _RETRY_DELAY,
     ) -> None:
-        """
+        """Create a new :class:`XpanderEventListener`.
+
         Parameters
         ----------
         api_key
@@ -79,6 +86,11 @@ class XpanderEventListener:
             Enable automatic LLM/execution metrics.
         max_sync_workers
             Size of the internal :class:`ThreadPoolExecutor` for sync handlers.
+        max_retries
+            How many times a failing request or SSE connection should be
+            retried before giving up.
+        retry_delay
+            Delay *in seconds* between retry attempts.
         """
         configure_git_credentials()
 
@@ -95,6 +107,8 @@ class XpanderEventListener:
         self.organization_id = organization_id
         self.base_url = xpander_client.configuration.base_url.rstrip("/")
         self.with_metrics_report = with_metrics_report
+        self.max_retries = max_retries
+        self.retry_delay = retry_delay
 
         # Internal resources
         self._pool: ThreadPoolExecutor = ThreadPoolExecutor(
@@ -106,7 +120,8 @@ class XpanderEventListener:
 
         logger.debug(
             f"XpanderEventListener initialised (base_url={self.base_url}, "
-            f"org_id={self.organization_id})"
+            f"org_id={self.organization_id}, retries={self.max_retries}, "
+            f"retry_delay={self.retry_delay}s)"
         )
 
     # --------------------------------------------------------------------- #
@@ -117,24 +132,10 @@ class XpanderEventListener:
         self,
         on_execution_request: ExecutionRequestHandler,
     ) -> None:
-        """
-        Start listening indefinitely (until cancelled or SIGINT/SIGTERM).
-
-        **Typical async usage**
-
-        ```python
-        async def handler(exec_: AgentExecution) -> AgentExecutionResult:
-            ...
-
-        listener = XpanderEventListener(...)
-        await listener.start(handler)          # inside an existing event-loop
-        ```
-
-        For synchronous code, see :pymeth:`register`.
-        """
+        """Start listening indefinitely (until cancelled or SIGINT/SIGTERM)."""
         loop = asyncio.get_running_loop()
 
-        # Graceful Ctrl-C / docker-stop / systemctl-stop
+        # Graceful Ctrl‑C / docker‑stop / systemctl‑stop
         for sig in (signal.SIGINT, signal.SIGTERM):
             loop.add_signal_handler(sig, lambda s=sig: asyncio.create_task(self.stop(s)))
 
@@ -153,11 +154,7 @@ class XpanderEventListener:
         await asyncio.gather(*self._bg)  # run forever (or until cancelled)
 
     async def stop(self, sig: signal.Signals | None = None) -> None:
-        """
-        Cancel background tasks and shut down the thread-pool.
-
-        Called automatically on SIGINT/SIGTERM but can also be invoked manually.
-        """
+        """Cancel background tasks and shut down the thread‑pool."""
         if sig:
             logger.info(f"Received {sig.name} – shutting down…")
 
@@ -170,7 +167,7 @@ class XpanderEventListener:
         self._bg.clear()
         logger.info("Listener stopped.")
 
-    # Pythonic context-manager sugar
+    # Pythonic context‑manager sugar
     async def __aenter__(self) -> "XpanderEventListener":
         return self
 
@@ -199,25 +196,57 @@ class XpanderEventListener:
     def _headers(self) -> dict[str, str]:
         return {"x-api-key": self.api_key}
 
+    # ---------------------- HTTP helpers with retry ---------------------- #
+
+    async def _request_with_retries(
+        self,
+        method: str,
+        url: str,
+        *,
+        headers: dict[str, str],
+        json: Any | None = None,
+    ) -> httpx.Response:
+        """Make an HTTP request with retry logic."""
+        last_exc: Exception | None = None
+        for attempt in range(1, self.max_retries + 1):
+            try:
+                async with httpx.AsyncClient() as client:
+                    response = await client.request(
+                        method,
+                        url,
+                        headers=headers,
+                        json=json,
+                        follow_redirects=True,
+                    )
+                return response
+            except Exception as exc:  # noqa: BLE001 broad
+                last_exc = exc
+                if attempt < self.max_retries:
+                    await asyncio.sleep(self.retry_delay)
+                else:
+                    logger.error(
+                        f"{method} {url} failed after {self.max_retries} attempts: {exc}"
+                    )
+        assert last_exc is not None  # for mypy
+        raise last_exc
+
     async def _release_worker(self, worker_id: str) -> None:
         url = f"{self._events_base()}/{worker_id}"
-        async with httpx.AsyncClient() as client:
-            await client.post(
-                url,
-                headers=self._headers(),
-                json=WorkerFinishedEvent().model_dump_safe(),
-                follow_redirects=True,
-            )
+        await self._request_with_retries(
+            "POST",
+            url,
+            headers=self._headers(),
+            json=WorkerFinishedEvent().model_dump_safe(),
+        )
 
     async def _make_heartbeat(self, worker_id: str) -> None:
         url = f"{self._events_base()}/{worker_id}"
-        async with httpx.AsyncClient() as client:
-            await client.post(
-                url,
-                headers=self._headers(),
-                json=WorkerHeartbeat().model_dump_safe(),
-                follow_redirects=True,
-            )
+        await self._request_with_retries(
+            "POST",
+            url,
+            headers=self._headers(),
+            json=WorkerHeartbeat().model_dump_safe(),
+        )
 
     async def _update_execution_result(
         self,
@@ -226,36 +255,58 @@ class XpanderEventListener:
     ) -> None:
         base = self._events_base().replace("/events", "/agent-execution")
         url = f"{base}/{execution_id}/finish"
-
-        async with httpx.AsyncClient() as client:
-            await client.patch(
-                url,
-                headers=self._headers(),
-                json={
-                    "result": execution_result.result,
-                    "status": (
-                        AgentExecutionStatus.Completed
-                        if execution_result.is_success
-                        else AgentExecutionStatus.Error
-                    ),
-                },
-                follow_redirects=True,
-            )
+        await self._request_with_retries(
+            "PATCH",
+            url,
+            headers=self._headers(),
+            json={
+                "result": execution_result.result,
+                "status": (
+                    AgentExecutionStatus.Completed
+                    if execution_result.is_success
+                    else AgentExecutionStatus.Error
+                ),
+            },
+        )
 
     async def _mark_execution_as_executing(self, execution_id: str) -> None:
         base = self._events_base().replace("/events", "/agent-execution")
         url = f"{base}/{execution_id}/finish"
+        await self._request_with_retries(
+            "PATCH",
+            url,
+            headers=self._headers(),
+            json={
+                "result": "",
+                "status": AgentExecutionStatus.Executing.value.lower(),
+            },
+        )
 
-        async with httpx.AsyncClient() as client:
-            await client.patch(
-                url,
-                headers=self._headers(),
-                json={
-                    "result": "",
-                    "status": AgentExecutionStatus.Executing.value.lower(),
-                },
-                follow_redirects=True,
-            )
+    # ----------------------- SSE helpers with retry ---------------------- #
+
+    async def _sse_events_with_retries(self, url: str):
+        """Yield SSE events with retry logic."""
+        attempt = 1
+        while True:
+            try:
+                async for event in aiosseclient(url=url, headers=self._headers(), raise_for_status=True):
+                    yield event
+                # If the stream ends gracefully, restart it after a short delay
+                logger.warning("SSE stream closed – reconnecting…")
+                await asyncio.sleep(self.retry_delay)
+                attempt = 1
+            except Exception as exc:  # noqa: BLE001 broad
+                if attempt >= self.max_retries:
+                    logger.error(
+                        f"SSE connection to {url} failed after {self.max_retries} attempts: {exc}"
+                    )
+                    raise
+                logger.warning(
+                    f"SSE connection to {url} failed (attempt {attempt}/{self.max_retries}) – "
+                    f"retrying in {self.retry_delay}s…"
+                )
+                attempt += 1
+                await asyncio.sleep(self.retry_delay)
 
     # --------------------------------------------------------------------- #
     # Internal helpers – execution path                                     #
@@ -279,7 +330,7 @@ class XpanderEventListener:
                     on_execution_request,
                     execution_task,
                 )
-        except Exception as exc:
+        except Exception as exc:  # noqa: BLE001 broad
             logger.exception("Execution handler failed")
             result.is_success = False
             result.result = f"Error: {exc}"
@@ -295,7 +346,7 @@ class XpanderEventListener:
         assert self._root_worker, "Root worker must be registered first"
         url = f"{self._events_base()}/{self._root_worker.id}/{agent_id}"
 
-        async for event in aiosseclient(url=url, headers=self._headers(), raise_for_status=True):
+        async for event in self._sse_events_with_retries(url):
             if event.event == EventType.WorkerRegistration:
                 agent_worker = DeployedAsset(**json.loads(event.data))
                 logger.info(f"Worker registered – id={agent_worker.id}")
@@ -307,8 +358,10 @@ class XpanderEventListener:
                     chat_url = f"https://{agent_meta.get('unique_name', agent_id)}.agents"
                     chat_url += ".stg" if is_stg else ""
                     chat_url += ".xpander.ai"
+                    
+                    builder_url = "https://"+("stg." if is_stg else "")+f"app.xpander.ai/agents/{agent_id}"
                     logger.info(
-                        f"Agent '{agent_meta.get('name', agent_id)}' chat: {chat_url}"
+                        f"Agent '{agent_meta.get('name', agent_id)}' chat: {chat_url} | builder: {builder_url}"
                     )
 
                 self._track(asyncio.create_task(self._heartbeat_loop(agent_worker.id)))
@@ -326,10 +379,9 @@ class XpanderEventListener:
     async def _register_parent_worker(self) -> DeployedAsset:
         url = self._events_base()
 
-        async for event in aiosseclient(url=url, headers=self._headers(), raise_for_status=True):
+        async for event in self._sse_events_with_retries(url):
             if event.event == EventType.WorkerRegistration:
-                root = DeployedAsset(**json.loads(event.data))
-                return root
+                return DeployedAsset(**json.loads(event.data))
 
         raise RuntimeError("Failed to register root worker – no WorkerRegistration received")
 
@@ -338,16 +390,18 @@ class XpanderEventListener:
     # --------------------------------------------------------------------- #
 
     def _track(self, task: asyncio.Task) -> None:
-        """Add *task* to the background set and auto-remove on completion."""
+        """Add *task* to the background set and auto‑remove on completion."""
         self._bg.add(task)
         task.add_done_callback(self._bg.discard)
 
     async def _heartbeat_loop(self, worker_id: str) -> None:
+        """Continuously send heartbeat events with retry logic."""
         while True:
             try:
                 await self._make_heartbeat(worker_id)
-            except Exception as exc:
-                pass # silent
+            except Exception as exc:  # noqa: BLE001 broad
+                pass
+                # logger.warning(f"Heartbeat for worker {worker_id} failed: {exc}")
             await asyncio.sleep(2)
 
     # --------------------------------------------------------------------- #
@@ -355,20 +409,5 @@ class XpanderEventListener:
     # --------------------------------------------------------------------- #
 
     def register(self, on_execution_request: ExecutionRequestHandler) -> None:
-        """
-        Start the listener in environments **without** an existing event-loop.
-
-        This helper spins up an internal loop via :func:`asyncio.run` and blocks
-        the current thread until the listener is stopped (e.g. with *Ctrl-C*).
-
-        **Example**
-
-        ```python
-        def blocking_main():
-            listener = XpanderEventListener(api_key=..., agent_id=...)
-            listener.register(my_handler)
-        ```
-
-        For fully-async applications prefer calling :pymeth:`start` directly.
-        """
+        """Synchronous helper for environments without an existing event‑loop."""
         asyncio.run(self.start(on_execution_request))
